package Games::BinaryishTrees::MultiNode;
use strict;
use warnings;

sub new {
  my $self = {};
  bless $self, shift;
#  $self->{top} = Games::BinaryishTrees::MultiNode::Node->new();
  $self->{top} = Games::BinaryishTrees::MultiNode::Node->new(undef, 0, shift); # so top has id utils::getroot() (so debug print in commented code avoids warning)
  return $self;
}

sub DESTROY {
  my $self = shift;
  $self->{top}->_clearrefs() if $self->{top};
}

1;
################################################################################
package Games::BinaryishTrees::MultiNode::Node;
use strict;
use warnings;

sub new {
  my $self = {};
  bless $self, shift;
  $self->{children} = [];
#  ($self->{parent}, $self->{counter}, $self->{id}, $self->{key}, $self->{value}) = @_;
  ($self->{parent}, $self->{id}, $self->{key}, $self->{value}) = @_;
  return $self;
}

sub _clearrefs {
  my $self = shift;

  for (@{$self->{children}}) {
    $_->_clearrefs() if defined $_;
  }

  delete $self->{children};
}

1;
################################################################################
package Games::BinaryishTrees::MultiNode::Handle;
use strict;
use warnings;

=head2 Tree::MultiNode::Handle::New

Constructs a new handle.  You must pass a tree object to Handle::New.

  my $tree   = new Tree::MultiNode;
  my $handle = new Tree::MultiNode::Handle($tree);

=cut

sub new {
  my $self = {};
  bless $self, shift;
  my $data = shift;
  $self->{tree} = $data;
  $self->{curr_node} = $data->{top};
  return $self;
}

=head2 Tree::MultiNode::Handle::get_data

Retrieves both the key, and value for the current node.

  my ($key,$val) = $handle->get_data();

=cut

sub get_data {
  my $self = shift;
  my $node = $self->{curr_node};

  return $node->{key}, $node->{value};
}

=head2 Tree::MultiNode::Handle::add_child

This member adds a new child node to the end of the array of children for the
current node.

  $handle->add_child();
  $handle->add_child($counter, $id, $key, $value);

=cut

#sub add_child {
#  my $self = shift;
#  my $child = Games::BinaryishTrees::MultiNode::Node->new(@_);
#  $child->{parent} = $self->{curr_node};
#  push @{$self->{curr_node}->{children}}, $child;
#}

=head2 Tree::MultiNode::Handle::top

Resets the handle to point back at the top of the tree.  
When top() is called, the current child becomes invalid (undef).

  $handle->top();

=cut

sub top {
  my $self = shift;
  $self->{curr_node} = $self->{tree}->{top};
  return 1;
}

sub selectid {
  my ($self, $key) = @_;

  for my $child (@{$self->{curr_node}->{children}}) {
    if ($key eq $child->{id}) {
      $self->{curr_node} = $child;
      return 1;
    }
  }

  return 0;
}

sub dumptree {
  my ($self, $place) = @_;
  if (defined $place && $place eq 'top') {
    _dumptree($self->{curr_node}->{children}->[0], 0);
  } else {
    _dumptree($self->{curr_node}, 0);
  }
}

sub _dumptree {
  my ($curr_node, $level) = @_;
  print ' ' x $level, "$curr_node->{id} ", defined $curr_node->{value} ? $curr_node->{value} : '', "\n";
#  print ' ' x $level, "$curr_node->{id}, $curr_node->{counter}, ", defined $curr_node->{value} ? $curr_node->{value} : '', "\n";
  $level++;

  for (@{$curr_node->{children}}) {
    _dumptree($_, $level);
  }

}

=head1 AUTHORS

This is a cut-down and customised version of Tree::Multinode. Tree::Multinode was written by 
Kyle R. Burton mortis@voicenet.com (initial version, and maintenence)
Daniel X. Pape dpape@canis.uiuc.edu (see Changes file from the source archive)
Eric Joanis <joanis@cs.toronto.edu>

=cut

1;
