package Games::BinaryishTrees::init;

use strict;
use warnings;
use Carp;
use Games::BinaryishTrees::messages qw(message);
use Games::BinaryishTrees::pairs qw(checkpairs);
#use Data::Dumper;
#use Data::Printer;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(initsolve findkos findpoortrees);

sub initsolve {
  my ($self) = @_;
#  my ($self, $treeref) = @_;
#  my ($self, @trees) = @_;
#  my @trees = @{$treeref};
#  croak message(18) if !@{$self->{trees}};
#  $self->{trees} = $treeref;
  checkpairs($self);
  initsomethings($self);
  changehistory($self);
  $self->{kos} = findkos($self);
  $self->{pp} = findpoortrees($self);
#  if (!$self->{loopflag}) {
#    $self->{kos} = {};
#  }
#  $self->{history}->{join '', @{$treeref}} = 1;
  $self->{history}->{join ':', @{$self->{trees}}} = 1;
}

#sub filltwigs {
#  my $self = shift;
#}

sub initsomethings {
  my $self = shift;
  $self->{gamebv} = $self->{bestvalue} = undef;
  $self->{kos} = {};
  $self->{pp} = {};
  $self->{bestpaths} = [];
  $self->{history} = {};
#  $self->{plays} = {}; # consider pooks 
  $self->{leaves} = {};
  $self->{mnTree} = {};
  $self->{mnhandle} = {};
}

sub changehistory {
  my $self = shift;
  my %h = %{$self->{prehistory}};
  $self->{history} = \%h;
}

sub findkos {
  my $self = shift;
  my $pairs = $self->{pairs};
  my @keyidscopy = keys %{$pairs};
  my %kos;

  for my $id (@keyidscopy) {

    for my $id2 (@keyidscopy) {
      if ($id eq $pairs->{$id2}->[0] || $id eq $pairs->{$id2}->[1]) {
        if ($pairs->{$id}->[0] eq $id2 || $pairs->{$id}->[1] eq $id2 ) {
          $kos{$id, $id2} = 1;
#          $kos{$id, $id2} = {$id => $id2, $id2 => $id};
        }
      }
    }

  }

  return \%kos;
}

sub findpoortrees { # looks through all pairs - its simpler than using $self->{trees} and finding options
  my $self = shift;
  my %pp;# = ('' => 1);
#  if ($self->{passflag} && !$self->{passplusflag}) {
  my $sflag;
  if (defined($self->{strategyvalue})) {
    $sflag = $self->{strategyvalue};
  } else {
    $sflag = 0;
  }
  if ($self->{passflag} && $self->{passplusflag}) {
#    my @poor;
    my @duffers;
    if ($self->{sweetflag} || ($sflag == 1)) { # sweetflag is as yet always false # this is not the case
#      @poor = grep {assesstree($self, 1, $_) == assesstree($self, -1, $_)} keys %{$self->{pairs}};
      @duffers = grep {assesstree($self, 1, $_) > assesstree($self, -1, $_)} keys %{$self->{pairs}}; # P:P alternative, this finds kos 
#    } elsif ($self->{strategyvalue} == 2) {
    } elsif ($sflag == 2) {
#      @poor = grep {assesstree($self, 1, $_, 1) + assesstree($self, -1, $_, 1) <= 0} keys %{$self->{pairs}};
#      @poor = grep {assesstree($self, 1, $_, 1) < 1 || assesstree($self, -1, $_, 1) < 1} keys %{$self->{pairs}};
#      @duffers = grep {assesstree($self, 1, $_, 1) > 0 || assesstree($self, -1, $_, 1) > 0} keys %{$self->{pairs}};

      for (keys %{$self->{pairs}}) {
        my $assessleft = assesstree($self, 1, $_, 1);
        my $assessright = assesstree($self, -1, $_, 1);
#        if ($assessleft < 0 || $assessright < 0 || ($assessleft == 0 && $assessright == 0)) {
#        if ($assessleft <= 0 && $assessright <= 0) {
#        if ($assessleft > 0 || $assessright > 0 || ($assessleft == 0 && $assessright == 0)) {
        if (($assessleft <= 0 && $assessright <= 0) || ($assessleft != $assessright)) {
          push @duffers, $_;
        }
      }

#      @duffers = grep {assesstree($self, 1, $_, 1) != assesstree($self, -1, $_, 1)} keys %{$self->{pairs}};
#      @duffers = keys %{$self->{pairs}}; # hoho
    } else {
      @duffers = grep {assesstree($self, 1, $_) <= assesstree($self, -1, $_)} keys %{$self->{pairs}};
    }
  $pp{$_} = 1 for @duffers;
  }
#print Dumper \%pp;
#p %pp;
  return \%pp;
}

sub assesstree {
  my ($self, $toggle, $tree, $addflag) = @_;
  my $score = 0;
#  print "$tree\n";
  my $looptracker = {};
  my @path = ();

  while ($tree) {
#    print "in while $tree\n"
#    carp "in while $tree\n";
    if (exists $looptracker->{$tree}) {
#      print "exiting ", join(':', @path), "\n";
      $self->{kos}->{$path[0], $path[1]} = 1; # we assume $path[1] exists - isok -> findkos finds loops of length 1
      $self->{kos}->{$path[1], $path[0]} = 1;
#      $self->{kos}->{$path[0], $path[1]} = {$path[0] => $path[1], $path[1] => $path[0]};
#      $self->{kos}->{$path[1], $path[0]} = {$path[0] => $path[1], $path[1] => $path[0]};
      last;
    }
    $looptracker->{$tree} = 1;
    push @path, $tree;
    ($tree, my $newscore, my $ko) = $self->_play($toggle, $tree, 1, $addflag);
#    ($tree, my $newscore, my $ko) = $self->_play($toggle, $tree, "$toggle$move"); # pooks call findpoortrees
#    print "here $tree, $ko\n";
    $score += $newscore;
    last if $ko;
    $toggle *= -1;
  }

  return $score;
}

=aftergooduse
sub assesstree {
  my ($self, $toggle, $tree) = @_;
  my $score = 0;

  while ($tree) {
    ($tree, my $newscore, my $ko) = $self->_play($toggle, $tree);
    $score += $newscore;
    last if $ko;
    $toggle *= -1;
  }

  return $score;
}
=cut

sub assesstree_notreally {
  my ($self, $toggle, $tree) = @_;
  my $score = 0;
#my $sillycounter = 0;
#print "first $tree\n";
  while ($tree) {
#  while ($tree || $sillycounter < 20) {
#  while ($sillycounter < 20) {
    ($tree, my $newscore, my $ko) = $self->_play($toggle, $tree);
#print "$tree/n" if $counter < 20;
#print "$tree/n";
#print "$tree\n";
#$sillycounter++;
# trouble reading error message as here to fix a+b<0
# well now see here to check comment about workaround
# for longway onemove recursion with passing and loop

#
#print "still here\n";
    $score += $newscore;
    last if $ko;
    $toggle *= -1;
  }

  return $score;
}

1;
