package Games::BinaryishTrees::misc;

use strict;
use warnings;
use Carp;
use Games::BinaryishTrees::utils qw(getfirst);
use Games::BinaryishTrees::messages qw(message);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(diff compare sente);

sub compare {
  my ($self, $trees, @params) = @_;
  my @store = store($self); 
  $self->setparams(@params);
  my @matches;
  my @treecopy1 = @{$trees};
  my @treecopy2 = @{$trees};

  for my $i (0..$#treecopy1) {
    my $t = shift @treecopy1;
    $matches[$i][$i] = "3";

    for my $j ($i..$#treecopy2) {
      if ($i != $j) {
        my $res = _diff($self, [$t, $treecopy2[$j]]);
        $matches[$i][$j] = "$res";
        $res =~ tr/54/45/;
        $matches[$j][$i] = "$res";
      }
    }

  }

  restore($self, @store);
  return @matches;
}

sub sente {
  my ($self, $tree, @params) = @_;
  my $sente = 0;
  my @store = store($self);
  $self->setparams(@params);

  for (0, 1) {
    $self->setstart($_);
#    (undef, my $bestvalue) = $self->_solve($tree);
#    my $game = $self->solve($tree);
    $sente += ($_ ? -1 : 1) * $self->solve($tree)->bestvalue;
  }

  restore($self, @store);
  return $sente;
}

sub diff {
  my ($self, $tree, @params) = @_;
  croak message(82) if ref($tree) ne 'ARRAY';
  croak message(83) if scalar @{$tree} != 2;
#  croak 'First parameter of "diff" needs to be an array reference' if ref($tree) ne 'ARRAY';
#  croak 'need two trees to diff and no more' if scalar @{$tree} != 2;
  my @store = store($self);
  $self->setparams(@params);
  my $string = _diff($self, $tree);
  restore($self, @store);
  return $string;
}

sub _diff {
  my ($self, $tree) = @_;
  my @choices;
  my $startwas = $self->setstart();

  for (0, 1) {
    $self->setstart($_);
#    (undef, undef, my $bestpaths) = $self->_solve($tree);
    $choices[$_] = [choicesbp($self->{trees}, $self->solve($tree)->bestpaths)];
  }

  $self->setstart($startwas);
  return parsechoices(@choices);
}

sub choicesbp {
  my ($trees, @bestpaths) = @_;
  my %cs = map { $_ => undef } getfirst(\@bestpaths);
  my @results;

  for my $letter (keys %cs) {
    for (0..$#{$trees}) {
      if ($trees->[$_] eq $letter) {
        push @results, $_;
        last;
      }
    }
  }

  return @results;
}

sub parsechoices {
  my @choices = @_;
  my $sc1 = scalar @{$choices[0]};
  my $sc2 = scalar @{$choices[1]};
  my $res;
  # $sc1 or $sc2 == 0 indicates that P (pass) was the only choice
  if (($sc1 == 0 && $sc2 == 0) || ($sc1 == 2 && $sc2 == 0) || ($sc1 == 0 && $sc2 == 2)) {
    $res = 3;
  } elsif ($sc1 == 0 && $sc2 == 1) {
    $res = $choices[1][0];
  } elsif ($sc1 == 1 && $sc2 == 0) {
    $res = $choices[0][0];     
  } elsif ($sc1 == 1 && $sc2 == 1) {
    if ($choices[0][0] eq $choices[1][0]) {
      $res = $choices[0][0] # the better choice
    } else {
#      $res = "2($choices[0][0])" # players choose differently
      $res = 2 # players choose differently
    }
  } elsif ($sc1 == $sc2) {
    $res = 3 # neither is preferred
  } else {
    my $pref = $sc1 < $sc2 ? $choices[0][0] : $choices[1][0];
    $res = $pref ? 5 : 4; # the second : first is at least as good as the first : second
  }
  return $res;
}

sub store {
  my $self = shift;
  my @params = $self->setparams();
  my @bp = defined $self->{bestpaths} ? @{$self->{bestpaths}} : ();
  return ($self->setstart(), $self->{bestvalue}, \@params, \@bp);
}

sub restore {
  my ($self, $start, $bv, $aparams, $abp) = @_;
  $self->setstart($start);
  $self->{bestvalue} = $bv;
  $self->setparams(@{$aparams});
  $self->{bestpaths} = $abp;
}

1;
