package Games::BinaryishTrees::playgame;

use strict;
use warnings;
use Games::BinaryishTrees::utils qw(getroot droproot getfirst indexof clone);
use Games::BinaryishTrees::messages qw(message notsoproudfunction aboveandbelow rightorleftmessage);
use Games::BinaryishTrees::solver qw(_play);
use Games::BinaryishTrees::viewer qw(viewpairtiny);
use Carp;
#use Data::Dumper;
#use Data::Printer;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(playgame newgame moveto weight moves bestchoices allchoices legalchoices tree gamepaths playout viewtree);

sub playgame {
  my $self = shift;
#  my $firstmoveflag = shift;
  _moveto($self, $self->{mnhandle}, $self->{game});
  my ($in, $out) = setinout($self, $self->{testing});
#  gameloop($self, $firstmoveflag, $in, $out, @_);
  gameloop($self, $in, $out, @_);
  if ($self->{testing}) {
    close $in;
    close $out;
  }
#  my @return = (score($self), moves($self));
  my @return = (weight($self), moves($self));
  newgame($self);
  return @return;
}

sub setinout {
  my ($self, $testing) = @_;
  my ($in, $out);
#  my $answersfolder = './answers';
  my $answersfolder = $self->{checkfolder};
  if ($testing) {
    open $in, '<', "$answersfolder/userinput.txt" or croak message(6);
    open $out, '>', "$answersfolder/clobber.txt" or croak message(7);
#    open $in, '<', './answers/userinput.txt' || croak message(6);
#    open $out, '>', './answers/clobber.txt' || croak message(7);
  } else {
    $in  = *STDIN;
    $out = *STDOUT;
  }
  return $in, $out;
}

sub gameloop {
  my ($self, $in, $out, @options) = @_;
#  my ($self, $firstmoveflag, $in, $out, @options) = @_;
#  my $sayagainnote = pop @_;
#  my $sayagainnote = defined $_->[10] ? $_->[10] : 1;

#  my $reasonableinput = 0;
#  my $firstmoveflag = $options[0];
  my $sayagainnote = defined $options[6] ? $options[6] : 1;
  my $firstmoveflag = $options[0];
#print @options, "\n";
  while (1) {
    my $userinput;
    if ($firstmoveflag) {
      $userinput = '';
      $firstmoveflag = undef;
    } else {
      print $out playoptions($self, @options);
      $userinput = <$in>;
      if (defined $userinput) {
        chomp $userinput;
  #      print $out message(8), "$userinput\n";
  #      print $out message(8), "$userinput\n" if $self->{testing}; # is idea anyway
#        print $out message(8), "$userinput\n" if $sayagainnote;
      } else {
#        print $out message(40, message(65));
#        $userinput = message(65);
        $userinput = "\4";
      }
      $userinput = message(65, $1) if $userinput =~ /([\cA-\cZ])/;
#      print $out message(8), "$userinput\n" if $sayagainnote;
      print $out message(8), topass($userinput, 51), "\n" if $sayagainnote;
    }
    my ($gameover, $string) = pgoutput(play($self, $userinput));
#    print $out $string, "\n";
    print $out $string, "\n" if $string;
    last if $gameover;
  }

}

sub playoptions {
  my ($self, undef, $bflag, $aflag, $lflag, $sflag, $mflag, undef, $vflag) = @_; # undef $sayagainnote
#print @_, "\n";
  my $string = '';
  my @tree;
  if ($aflag || $vflag) {
    @tree = allchoices($self);
  }
  $string .= stringify(message(0), bestchoices($self)) if $bflag;
  $string .= stringify(message(1), @tree) if $aflag;
#  $string .= stringify(message(1), allchoices($self)) if $aflag;
  $string .= stringify(message(2), legalchoices($self)) if $lflag;
#  $string .= stringify(message(3), score($self)) if $sflag;
#  $string .= stringify(message(3), notsoproudfunction(score($self))) if $sflag;
#  $string .= stringify(message(3), notsoproudfunction(weigh($self))) if $sflag;
  $string .= stringify(message(3), weight($self)) if $sflag;
#  $string .= stringify(message(3), notsoproudfunction(my $ditchthis, score($self))) if $sflag;
  $string .= stringify(message(4), moves($self)) if $mflag;
#  $string .= join '', message(5, rightorleftmessage()->{(($self->{startflag} + movenumber($self->{game})) % 2) ? -1 : 1});
#$string = join '', $string, $self->{gbt}->viewpair('b'), "\n";
#$string = join '', $string, Games::BinaryishTrees::viewpair($self, 'b'), "\n";
#print " $vflag, vflags ", @vflags, "\n";

#  $string .= join '', Games::BinaryishTrees::view($self, \@vflags), "\n" if $vflag;
  $string .= viewtree($self, @tree) if $vflag;
#p($self->{gbt}->{pairs});
#p($self->{pairs});
  $string .= message(5, rightorleftmessage()->{(($self->{startflag} + movenumber($self->{game})) % 2) ? -1 : 1});
  return $string;
}

sub viewtree {
  my ($self, @tree) = @_;
  my $string = join '', map { viewpairtiny($self, $_) } grep { $_ ne 'P' } @tree;
}

sub stringify {
  my $message = shift;
#  join(' ', "$message =", @_, "\n"); # make message for =
#  join(' ', $message.message(63), @_, "\n"); # make message for =
#  join(' ', $message.message(65).message(63), @_, "\n"); # make message for =
#  join(' ', $message.message(63), @_, "\n"); # make message for  = # tiredcomment
#  join('*', $message, message(63), @_, "\n"); # make message for  = 
#  join(' ', $message, message(63), @_, "\n"); # make message for  = 
#  join(' ', $message, message(43), @_, "\n"); # put = in
  join(message(80), $message, message(43), @_, "\n"); # put = in
}

sub pgoutput {
  my ($answer, $errorcode, $errorstring) = @_;
  my $string;
  my $gameover = 0;
  if ($errorcode) {
    if ($answer =~ /^(-1)$|^[a-zA-Z]/) {
      $string = message(10).$answer."\n" if !$1;
      $string .= message(9);
      $gameover = 1;
    } else {
      $string = $errorstring;
    }
  } else {
    if (defined $answer) { # todo two player version # todone
      if (substr($answer, 0, 1) ne '-') {
        if ($answer) {
          $answer = $answer eq 'P' ? message(44) : $answer; 
          $string = message(10).$answer;# if $answer;
        }
      } else {
        $string = message(11).$answer;
      }
    }
  }
#  return $gameover, $string; # <littlecomment>$string =~ s/P/message()/eg</littlecomment>
#  return $gameover, $string; # <littlecomment>$string =~ s/P/message()/eg</littlecomment>
#  $string =~ s/P/message()/eg; # another before look 
#  $string =~ s/P/message(64)/eg; # another before look
#  $string = $string eq 'P' ? message(64) : $string;
#  $string = $string eq 'P' ? message(44) : $string; 
  return $gameover, $string; # returnmissing
}

sub allchoices {
  my $self = shift;
#  my @notempty = grep { length($_) } tree($self);
#  my @notempty = grep { length($_) } tree($_[0]);
#  push @notempty, 'P'; # make message for P
#  push @notempty, message(65); # make message for P
  my @tree = tree($self);
#  push @tree, message(45) unless inarray(\@tree, 'P');
  my $indexofP = indexof('P', @tree);
  if (defined $indexofP) {
    $tree[$indexofP] = message(45);
  } else {
    push @tree, message(45);
  }
  return sort @tree;
}

sub legalchoices {
  my $self = shift;
#  return sort map { $_->{id} } _moveto($self, $self->{mnhandle}, $self->{game});
#  return sort map { topass($_->{id}, 28) } _moveto({}, $self->{mnhandle}, $self->{game}); # looks safer but only
  return sort map { topass($_->{id}, 28) } _moveto($self, $self->{mnhandle}, $self->{game}); # looks safer but only
}

=goodbreak comment
sub score {
  my $self = shift;
  if (defined $self->{game}) {
    if (defined $self->{gamebv} && defined $self->{bestvalue}) {
      my $score = $self->{bestvalue} - $self->{gamebv};
      $score = -$score if $self->{startflag};
#      return $score > 0 ? "+$score" : $score; # make message for +
      return $score > 0 ? message(62).$score : $score; # make message for + # i do big cheat look i read name so back to simple way
    } else {
      return message(12);
    }
  } else {
    return message(13), "\n";
  }
}
=cut

sub weight {
  my $self = shift;
  if (defined $self->{game}) {
    if (defined $self->{gamebv} && defined $self->{bestvalue}) {
      my $score = $self->{bestvalue} - $self->{gamebv};
      $score = -$score if $self->{startflag};
#      return $score > 0 ? "+$score" : $score; # make message for +
#      return $score > 0 ? message(62).$score : $score; # make message for + # i do big cheat look i read name so back to simple way
#      return $score == 0 ? message(66) : $score > 0 ? message(62).$score : $score;
#      if ($score == 0) {
#        return message(66);
#        return message(46);
#      } else {  
#        my $sign = $score > 0 ? message(62) : '';
#        my $sign = $score > 0 ? message(42) : '';
#        my $sign = $score > 0 ? message(42) : $score < 0 ? message(66) : '=';
        my $sign = $score > 0 ? message(42) : $score < 0 ? message(66) : '=';
        if ($self->{changeallnumbersflag}) {
#          return $sign.aboveandbelow($self->{aboveandbelow}, $score);
#          return aboveandbelow($self->{aboveandbelow}, $score);
#          return join '', $sign, abs aboveandbelow($self->{aboveandbelow}, $score);
#          return join '', $sign, absnumber(aboveandbelow($self->{aboveandbelow}, $score));
#          if ($sign eq '=') {
          return join '', aboveandbelow($self->{aboveandbelow}, $score);
#          } else {
#            return join '', $sign, absnumber(aboveandbelow($self->{aboveandbelow}, $score)); #haha
#          }
        } else {
          if ($score == 0) {
            return message(46);  
          } else {
#print "weight score $score\n";
#            return join '', $sign, absnumber(notsoproudfunction($score));
#            return join '', $sign, notsoproudfunction(absnumber($score));
            return join '', $sign, notsoproudfunction(abs($score));
          }
        }
#      }
    } else {
      return message(12);
    }
  } else {
    return message(13); # ? to recieve this one
  }
}

sub newgame {
  my $self = shift;
  $self->{game} = $self->{root};
  $self->{mnhandle}->top();
#  $self->{mnTree}->{game} = '';
}

sub moves {
  my $self = shift;
  return join ':', map { topass($_, 68) } split ':', droproot($self->{game});
}

sub movenumber {
  my $str = shift;
  my $count = $str =~ tr/://;
  return $count;
}

sub moveto {
  my ($self, $path) = @_;
  $path ||= '';
  $path = reversetopass(61, $path, $self);
  _moveto($self, $self->{mnhandle}, join(':', $self->{root}, $path)) >= 0 ? 1 : 0;
}

sub cleanupmoveto {
  my ($bp, $bv, $length) = @_;
  my $root = getroot();
  my @keys = map { "$root:$_" } split "$root:", $bp;
  shift @keys;
  @keys = map { $length + 1 > length($_) ? undef : substr($_, $length + 1) } @keys; 
  # length + 1 to remove : from front of substring, but as path does not end with a : grabbing everything leads to substr warning unless checked for
  return $bv, \@keys;
}

sub _moveto {
  my ($self, $mnhandle, $path) = @_;
  $mnhandle->top();
  my $counter = 0;
#p $self;
#p $self->{originalpairs};
  $self->{pairs} = clone($self->{originalpairs});
#print "_moveto $path\n";
  for my $id (split ':', $path) {
#print "_moveto id is $id\n";
#    return -1 unless $mnhandle->selectid($_);
    return -1 unless $mnhandle->selectid($id);
#print "past unless\n";
    if (exists $self->{pairs}->{$id}) {
      my $pairid = $self->{pairs}->{$id};
      playout($self, ($counter++ + $self->{startflag}) % 2, $id, $pairid);
    }
  }

  ($self->{bestvalue}, $self->{gamepaths}) = cleanupmoveto($mnhandle->get_data(), length($path));
  $self->{game} = $path;
  return @{$mnhandle->{curr_node}->{children}};
}

sub playout {
  my ($self, $toggle, $id, $pairid) = @_;
  my ($score, $move);
  my $rightorleft = $pairid->[$toggle];
  if (exists $self->{pairs}->{$rightorleft}) {
    $move = $rightorleft;
    if ($self->{loopflag} && exists $self->{kos}->{$id, $rightorleft}) {
      $score = $toggle;
    } else {
      $score = 0;
    }
  } else {
    $score = $toggle ? -$rightorleft : $rightorleft;
  }
  if (defined $pairid->[4]) {

#    while (my ($k, $v) = each %{$pairid->[4]}) {
    for my $k (keys %{$pairid->[4]}) {
      Games::BinaryishTrees::pairs::_pair($self, $k, $pairid->[4]->{$k});
#      Games::BinaryishTrees::pairs::_pair($self, $k, @v);
    }

  }
  return $score, $move;
}

sub undo {
  my ($self, $undos) = @_;
  my @moves = split ':', $self->{game};
  if (@moves > $undos) {
    pop @moves for 1..$undos;
    $self->{game} = join ':', @moves;
    return 1;
  } else {
    return 0;
  }
}

sub bestchoices {
  my $self = shift;
#print join("\n", @{$self->{gamepaths}}), "\n";
#  my %cs = map { topass($_, 48) => undef } getfirst($self->{gamepaths});
#p $self->{gamepaths};
#  my @cs = map { topass($_, 48) => 1 } getfirst($self->{gamepaths});
  my %cs = map { topass($_, 48) => 1 } getfirst($self->{gamepaths});
  return sort keys %cs;
}

sub topass {
  my ($pass, $passmessagenumber) = @_;
  if ($pass eq 'P') {
    return message($passmessagenumber);
  } else {
    return $pass;
  }
}

sub reversetopass {
  my ($messagenumber, $path, $self) = @_;
  my $message = message($messagenumber);
  my @string;

  for my $move (split ':', $path) {
    if ($move eq $message) {
      push @string, 'P';      
    } else {
      if ($move eq 'P') {
        carp message(81, message(61)) unless $self->{warningsoffflag};
        push @string, 'P'; 
      } else {
        push @string, $move;
      }
    }

  }

  return join ':', @string;
#  return join ':', map { $_ eq $message ? 'P' : $_ } split ':', $path;
}

sub tree {
  my $self = shift;
  my @okchoices = legalchoices($self);
#  my $repeats = $self->{gbt}->{pathrepeat};
  my $repeats = $self->{pathrepeat};

#print Dumper $forcedpasses;
  if (exists $repeats->{$self->{game}}) {
    push @okchoices, keys %{$repeats->{$self->{game}}};
  }

#p @okchoices;
#=incut
#  my @moves = split ':', droproot($self->{game});
#  if (@moves) {
#    my $lastmove = $moves[-1];
#    my $toggle = ($self->{startflag} + $#moves) % 2;

#    for my $key (keys %{$self->{kos}}) {
#      if (exists $self->{kos}->{$key}->{$lastmove} && $self->{pairs}->{$lastmove}->[$toggle] eq $self->{kos}->{$key}->{$lastmove}) {
#        push @okchoices, $self->{kos}->{$key}->{$lastmove};
#        last; # adds one at most, however two loops may cause two moves to be needed to be added here
#      }
#    }
  
#  }
#=cut

  return @okchoices;
}

sub play {
  my ($self, $move) = @_;
  if (length($move)) {
    if ($move eq message(16)) {
      return -1, 1, '';
    }
    if ($move =~ /^(-?\d+)$/) {
      return trytomoveback($self, -$1, $move);
    }
    if ($move eq message(51)) {
      $move = 'P';
    } else {
      if ($move =~ /^[a-zA-Z]/ && $move !~ /:/) {
      } else {
  #      return -2, 1, movewithvariable(0, $move);
  #      return -2, 1, messagewithvariable(0, $move);
  #      return -2, 1, messageaddsimple(message(58), $move, message(59));
  #      return -2, 1, messageaddsimple(40, $move);
        return -2, 1, message(40, $move);
      }
    }
#print "move is ($move)\n";
    $self->{game} .= ":$move"
  }
  return choosemove($self, $move);
}

sub trytomoveback {
  my ($self, $back, $move) = @_;
#  if ($back && undo($self, $back) && _moveto($self, $self->{mnhandle}, $self->{game}) >= 0) {
  if ($back > 0 && undo($self, $back) && _moveto($self, $self->{mnhandle}, $self->{game}) >= 0) {
    return -$back, 0, '';
  } else {
    return -3, 1, message(14).$move;
  }
}

sub choosemove {
  my ($self, $move) = @_;
  my $children = _moveto($self, $self->{mnhandle}, $self->{game});
  my @retval = ('', 0, '');
  if ($children > 0) {
    @retval = computermove($self) if $self->{computerplaysflag};
  } elsif (!$children) {
    @retval = (-1, 1, ''); # end of game # yes this one is forlookat
  } else {
    my @tmb = trytomoveback($self, 1, $move);
    if ($tmb[1]) {
      @retval = @tmb;
    } else {
#      @retval = (-4, 1, "move $move is not allowed\n");
#      @retval = (-4, 1, messagewithvariable(1, $move));
#      @retval = (-4, 1, messageaddsimple(message(60), $move, message(61)));
      @retval = (-4, 1, message(41, $move));
#      @retval = (-4, 1, message(41, topass($move, 44))); # either way hmm
    }
  }
  return @retval;
}

sub computermove {
  my $self = shift;
  my @answers = bestchoices($self);
  @answers = legalchoices($self) if !@answers; # maybe postprocessquicker is set and this part of the tree was abandoned
  croak move(15) unless @answers;
  my $answer = $answers[rand @answers];
  _moveto($self, $self->{mnhandle}, "$self->{game}:$answer");
  return ($answer, gameover($self->{game}), '');
}

sub gameover {
  substr($_[0], -3) eq 'P:P';
}

sub gamepaths {
  map { droproot(join(':', map { topass($_, 48) } split ':' )) } @{$_[0]->{gamepaths}};
#  map { join ':', map { topass($_, 48) } split ':' } @{$_[0]->{gamepaths}};
#  map { droproot($_) } @{$_[0]->{gamepaths}};
}

#sub AUTOLOAD {};

#sub changestupods($message, $index) {
#sub changestupods($index, $message) {
#sub changestupods {
#sub changestupod {
#  my $self = shift;
#  return changemessagefromindexforgame(@_);
#}

#sub changemessage {
#  my $self = shift;
##  return changemessage(@_);
###  return changemessage(@_); # simpleslipup
#  return changethismessage(@_);
#}

#sub changethisstupods {
#  my $self = shift;
#  return changemessage(@_);
#}

1;
