package Games::BinaryishTrees::twigless;

use strict;
use warnings;
use Games::BinaryishTrees::utils qw(indexof);
#use Data::Dumper;
#use Data::Printer;
require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(optleaves permleaves);

sub permleaves {
  my ($gbt, $path, $value, $toggle, $sflag, @tree) = @_;
#print "perming\n";
#p @tree;
  my $aperms = [];
  permute($aperms, '', @tree);
#  my $lmbflag = $gbt->{lmbflag} && (@tree % 2);
#  my $lmb = $sflag ? $gbt->{lmbflag} && (@tree % 2) : $toggle * ($gbt->{lmbflag} && (@tree % 2));
  my $lmb = $sflag ? $gbt->{lmbflag} * (scalar @tree) : $toggle * ($gbt->{lmbflag} && (@tree % 2));

  for (@$aperms) {
    my @moves = split ':';
    shift @moves; # remove root # turns out throwing away empty value
    $gbt->{leaves}->{"$path$_:P:P"} = $value + getvalue($gbt->{pairs}, $toggle, $sflag, @moves) + $lmb; 
  }

}

sub optleaves { # since optleaves sorts, doesn't work with sweetflag # and strategy(2)
  # (exits if there's a group of 2 or more with the same value because given s1 = s2 and l < s1,
  # s1:s2:l, s2:s1:l, l:s1:s2, l:s2:s1 could all be solutions)
  my ($gbt, $path, $value, $toggle, undef, @tree) = @_;
  my %hash = group($gbt->{twigless}, @tree);
#print "opting\n";
#print Dumper \%hash;
#print "tree\n", Dumper \@tree;
  my $pairs = $gbt->{pairs};

  for my $key (sort { $b <=> $a } keys %hash) {
    if (@{$hash{$key}} == 1) {
      my $move = $hash{$key}->[0];
      $path .= ":$move";
#      if ($sflag) {
#        $value += $pairs->{$move}->[$toggle < 0] + $gbt->{lmbflag};
#      } else {
        $value += ($toggle * $pairs->{$move}->[$toggle < 0]) + ($gbt->{lmbflag} ? $toggle : 0);
#      }
      $toggle = -$toggle;
      splice @tree, indexof($move, @tree), 1;
    } else {
      last;
    }
  }

  if (@tree) {
#    permleaves($gbt, $path, $value, $toggle, @tree);
#    permleaves($gbt, $path, $value, $toggle, $sflag, @tree);
    permleaves($gbt, $path, $value, $toggle, undef, @tree);
  } else {
    $gbt->{leaves}->{"$path:P:P"} = $value;
  }
}

sub getvalue {
  my ($pairs, $toggle, $sflag, @tree) = @_;
  my $value = 0;

  for my $move (@tree) {
    if ($sflag) {
      $value += $pairs->{$move}->[$toggle < 0];
    } else {
      $value += $toggle * $pairs->{$move}->[$toggle < 0];
    }
#    $value += $toggle * $pairs->{$_}->[$toggle < 0];
    $toggle = -$toggle;
  }

  return $value;
}

sub group {
  my ($twigless, @tree) = @_;
  my %hash;

  for (@tree) {
    push @{$hash{$twigless->{$_}}}, $_;
  }

  return %hash;
}

sub permute { # @array always has 2 or more elements because smaller arrays are filtered out at optleaves and solver
  my ($res, $thisstring, @array) = @_;
  if ($#array == 1) {
    push @$res, join(':', $thisstring, @array), join(':', $thisstring, reverse @array);
  } else {
  
    for my $index (0..$#array) {
      my @copy = @array;
      permute($res, join(':', $thisstring, splice @copy, $index, 1), @copy);
    }

  }
}

# code below puts P into twigless paths # slower 
# setprocess can give a big speedup (optleaves)
# passplus can give a big speedup when pass enabled

sub permthesePleaves {
  my ($gbt, $path, $value, $toggle, @tree) = @_;
#print "perming from $path\n";
#p @tree;
  my $aperms = [];
  permute($aperms, '', @tree);
  my @Ppaths = ([]);
  my @aperms = @{$aperms};
#p @aperms;

  for my $permpath (@aperms) {
    my @permpath = split ':', $permpath;
    shift @permpath; # first value empty

    for (0..$#permpath) {
      push @Ppaths, insertPs(@permpath[0..$_]);
    }

  }
#p @Ppaths;
  for (@Ppaths) {
#p $_;
#    my @moves = split ':';
    my $length = scalar @{$_};
    my $lmbflag = $gbt->{lmbflag} && ($length % 2);
    my $suffixpath = $length ? ':'.join(':', @$_).':' : ':'; # adding empty a few times ..
#p @$_;
#print "$path${suffixpath}P:P ", $value + getPvalue($gbt, $toggle, $lmbflag, @$_), "\n";
    $gbt->{leaves}->{"$path${suffixpath}P:P"} = $value + getPvalue($gbt, $toggle, $lmbflag, @$_);
  }

}

sub getPvalue {
  my ($gbt, $toggle, $lmbflag, @tree) = @_;
  my $pairs = $gbt->{pairs};
  my $value = 0;

  for (@tree) {
#print "move is $_\n";
    if ($_ eq 'P') {
      $value += -$toggle * $gbt->{passtaxflag};
    } else {
      $value += $toggle * $pairs->{$_}->[$toggle < 0];
    }
    $toggle = -$toggle;
  }

  $value += -$toggle if $lmbflag;# && (scalar @tree) % 2; # see permthesePleaves
  return $value;
}

sub insertPs {
  my @path = @_;
#  shift @path; # remove root
  my @newpaths;
  my $movecount = scalar @path;
  my $top = 2**($movecount) - 1;

  for (0..$top) {
    my $binary = dec2bin($movecount, $_);
    my $index = 0;
    my @pathcopy = @path;

    for (split '', $binary) {
      if ($_) {
        splice @pathcopy, $index, 0, 'P';
        $index++
      }
      $index++;
    }

    push @newpaths, \@pathcopy;

  }
  return @newpaths;
}

sub dec2bin { # adds a 0 at the end when not sent enough
  my $paramnum = shift;
  my $str = unpack('B32', pack('N', shift));
  my $remove = '0' x (32 - $paramnum);
  $str =~ s/^$remove//;
  return $str;
}

1;
