use strict;
use warnings;
use lib qw(../../lib);
use Games::BinaryishTrees;
use JSON::MaybeXS;

my ($game, $pathtotrees) = @ARGV;
$pathtotrees ||= './TreesJSON';
if (-d $pathtotrees && -e "$pathtotrees/simple.json") {
  my @games = Games::BinaryishTrees::filer::readfolder($pathtotrees);

  if (Games::BinaryishTrees::utils::inarray(\@games, $game)) {
    my @leftstartspaths;
    my @rightstartspaths;

    for my $setting (0..1) {
      my $gbt = new Games::BinaryishTrees;
      $gbt->setwarningsoff(1);
      my $pairset = $gbt->fromjson(decode_json($gbt->readfile("$pathtotrees/$game")));
      $gbt->setsweet(1); # set these after fromjson
      $gbt->setstart($setting);
      my $solution = $gbt->solve($pairset);
      if ($setting) {
        @rightstartspaths = $solution->bestpaths();
      } else {
        @leftstartspaths = $solution->bestpaths();
  #      @rightstartspaths = $solution->bestpaths();
      }
    }

    print compare(\@leftstartspaths, \@rightstartspaths), "\n";

  } else {
    print "usage : symmetryJSON.pl [", join(' ', @games), "] path/to/Trees\n\nCompares all best paths with their reverse when passed first\n";
  }
} else {
  print "useage : symmetryJSON.pl game path/to/TreesJSON\n";
}

sub compare {
  my ($aref, $bref) = @_;
  my %ahash = map { join(':', grep { $_ ne 'P' } split ':') => 1 } @$aref;
  my %bhash = map { join(':', grep { $_ ne 'P' } split ':') => 1 } @$bref;
  my %breversed;

  for my $bkey (keys %bhash) {
    my $reverse = join ':', reverse split ':', $bkey;
    $breversed{$reverse} = $bhash{$bkey};
  }

  return comparehashes(\%ahash, \%breversed);
}

sub comparehashes {
  my ($ahashref, $bhashref) = @_;
  if (keys %{$ahashref} == keys %{$bhashref}) {

    for my $akey (keys %{$ahashref}) {
      if (exists $bhashref->{$akey} && $ahashref->{$akey} eq $bhashref->{$akey}) {
      } else {
        return 'differing';
      }
    }

  } else {
    return 'differing';
  }
  return 'equal';
}

